from __future__ import division

from PIL import Image

from . import modes
from .transform import Transform


class ImageSize(object):

    @property
    def image(self):
        if not self._image and self.path:
            self._image = Image.open(self.path)
        return self._image

    def __init__(self, path=None, image=None, width=None, height=None,
        enlarge=True, mode=None, transform=None, sharpen=None, _shortcut=False, **kw
    ):

        # Inputs.
        self.__dict__.update(kw)
        self.path = path
        self._image = image
        self.req_width = width
        self.req_height = height
        self.enlarge = bool(enlarge)
        self.mode = mode
        self.transform = transform
        self.sharpen = sharpen

        self.image_width = self.image_height = None

        # Results to be updated as appropriate.
        self.needs_enlarge = None
        self.width = width
        self.height = height
        self.op_width = None
        self.op_height = None

        if _shortcut and width and height and enlarge and mode in (modes.RESHAPE, modes.CROP, None):
            return

        # Source the original image dimensions.
        if self.transform:
            self.image_width, self.image_height = Transform(self.transform,
                self.image.size if self.image else (width, height)
            ).size
        else:
            self.image_width, self.image_height = self.image.size

        # Maintain aspect ratio and scale width.
        if not self.height:
            self.needs_enlarge = self.width > self.image_width
            if not self.enlarge:
                self.width = min(self.width, self.image_width)
            self.height = self.image_height * self.width // self.image_width
            return

        # Maintain aspect ratio and scale height.
        if not self.width:
            self.needs_enlarge = self.height > self.image_height
            if not self.enlarge:
                self.height = min(self.height, self.image_height)
            self.width = self.image_width * self.height // self.image_height
            return

        # Don't maintain aspect ratio; enlarging is sloppy here.
        if self.mode in (modes.RESHAPE, None):
            self.needs_enlarge = self.width > self.image_width or self.height > self.image_height
            if not self.enlarge:
                self.width = min(self.width, self.image_width)
                self.height = min(self.height, self.image_height)
            return

        if self.mode not in (modes.FIT, modes.CROP, modes.PAD):
            raise ValueError('unknown mode %r' % self.mode)

        # This effectively gives us the dimensions of scaling to fit within or
        # around the requested size. These are always scaled to fit.
        fit, pre_crop = sorted([
            (self.req_width, self.image_height * self.req_width // self.image_width),
            (self.image_width * self.req_height // self.image_height, self.req_height)
        ])

        self.op_width, self.op_height = fit if self.mode in (modes.FIT, modes.PAD) else pre_crop
        self.needs_enlarge = self.op_width > self.image_width or self.op_height > self.image_height

        if self.needs_enlarge and not self.enlarge:
            self.op_width = min(self.op_width, self.image_width)
            self.op_height = min(self.op_height, self.image_height)
            if self.mode != modes.PAD:
                self.width = min(self.width, self.image_width)
                self.height = min(self.height, self.image_height)
            return

        if self.mode != modes.PAD:
            self.width = min(self.op_width, self.width)
            self.height = min(self.op_height, self.height)



from PIL import Image


TRANSFORM_AXIS = {
    Image.EXTENT: (0, 1, 0, 1),
    Image.AFFINE: (None, None, 0, None, None, 1),
    Image.QUAD: (0, 1, 0, 1, 0, 1, 0, 1),
    Image.PERSPECTIVE: (None, None, None, None, None, None, None, None),
    # Image.MESH: ???
}


class Transform(list):

    def __init__(self, spec, image_size=None):

        super(Transform, self).__init__(spec)

        self.flag = getattr(Image, self[0].upper())
        try:
            axis = (None, 0, 1) + TRANSFORM_AXIS[self.flag]
        except KeyError:
            raise ValueError('unknown transform %r' % self[0])

        if len(self) != len(axis):
            raise ValueError('expected %d transform values; got %d' % (len(axis), len(self)))

        for i in xrange(1, len(self)):
            v = self[i]
            if isinstance(v, basestring):
                if v[-1:] in ('%', 'p'): # Percentages.
                    if axis[i] is None:
                        raise ValueError('unknown dimension for %s value %d' % (self[0], i))
                    if image_size is None:
                        raise ValueError('no image size with relative transform')
                    self[i] = image_size[axis[i]] * float(v[:-1]) / 100
                else:
                    self[i] = float(v)

        # Finalize the size.
        if not self[1] or not self[2]:
            if not image_size:
                ValueError('no image size or transform size')
        self[1] = int(self[1] or image_size[0])
        self[2] = int(self[2] or image_size[1])

    @property
    def size(self):
        return self[1], self[2]

    def apply(self, image):
        return image.transform(
            (int(self[1] or image.size[0]), int(self[2] or image.size[1])),
            self.flag,
            self[3:],
            Image.BILINEAR,
        )



from __future__ import division

import os
import errno
from PIL import Image
from boto.s3.connection import S3Connection
from flask import current_app, request, abort, redirect
from urllib import quote as urlquote

from . import modes
from .exception import ParameterNotFound, FilterNotFound, OriginalKeyDoesNotExist
from .size import ImageSize
from .transform import Transform


def makedirs(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


class Imagine(object):
    filters = {}
    s3_conn = None
    bucket = None

    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        if not hasattr(app, 'extensions'):
            app.extensions = {}
        app.extensions['imagine'] = self

        app.config.setdefault('IMAGINE_URL', '/media/cache/resolve')
        app.config.setdefault('IMAGINE_NAME', 'imagine')
        app.config.setdefault('IMAGINE_TYPE', 's3') # s3 or filesystem
        app.config.setdefault('IMAGINE_THUMBS_PATH', 'cache')
        app.config.setdefault('IMAGINE_CACHE', '/tmp/flask-imagine')

        if app.config['IMAGINE_TYPE'] == 's3':
            if 'IMAGINE_S3_ACCESS_KEY' not in app.config \
                    or 'IMAGINE_S3_SECRET_KEY' not in app.config \
                    or 'IMAGINE_S3_BUCKET' not in app.config:
                raise ParameterNotFound(code=101, msg='S3 credentials has been not present')

        self.s3_conn = S3Connection(app.config['IMAGINE_S3_ACCESS_KEY'], app.config['IMAGINE_S3_SECRET_KEY'])
        self.bucket = self.s3_conn.get_bucket(app.config['IMAGINE_S3_BUCKET'])

        if 'IMAGINE_FILTERS' not in app.config:
            raise ParameterNotFound(code=102, msg='Filters configuration has been not present')

        self.filters = self._prepare_filters(app.config['IMAGINE_FILTERS'])

        app.add_url_rule(app.config['IMAGINE_URL'] + '/<regex("[^\/]+"):filter_name>/<path:path>', app.config['IMAGINE_NAME'], self.handle_request)

        if hasattr(app, 'add_template_filter'):
            app.add_template_filter(imagine_filter, 'imagine_filter')
        else:
            ctx = {
                'imagine_filter': imagine_filter
            }
            app.context_processor(lambda: ctx)

    @staticmethod
    def _prepare_filters(filters_config):
        """
        Filters validation
        :param filters_config: dict
        :return: dict
        """
        # todo: Must be realized in the future
        return filters_config

    def build_url(self, path, filter_name, **kwargs):
        if filter_name not in self.filters:
            raise FilterNotFound(code=201, msg='Filter %s has been not found in config' % filter_name)

        makedirs(current_app.config['IMAGINE_CACHE'])

        cached_key = self.bucket.get_key(current_app.config['IMAGINE_THUMBS_PATH'] + '/' + filter_name + '/' + path)

        if cached_key is not None:
            return cached_key.generate_url(expires_in=0, query_auth=False)
        else:
            original_key = self.bucket.get_key(path)
            if original_key is not None:
                external = kwargs.pop('external', None) or kwargs.pop('_external', None)
                scheme = kwargs.pop('scheme', None)
                if scheme and not external:
                    raise ValueError('cannot specify scheme without external=True')

                url = '%s/%s/%s' % (
                    current_app.config['IMAGINE_URL'],
                    urlquote(filter_name),
                    urlquote(path),
                )

                if external:
                    url = '%s://%s%s/%s' % (
                        scheme or request.scheme,
                        request.host,
                        request.script_root,
                        url.lstrip('/')
                    )

                return url
            elif 'default_image_path' in self.filters[filter_name]:
                return self.filters[filter_name]['default_image_path']
            else:
                raise OriginalKeyDoesNotExist(code=404,
                                              msg='Original key <%s> does not exist in bucket: %s' % (
                                                  path,
                                                  self.bucket.name
                                              ))

    def handle_request(self, filter_name, path):
        if filter_name not in self.filters:
            abort(404)
        original_key = self.bucket.get_key(path)

        if 'filter' in self.filters[filter_name]:
            if self.filters[filter_name]['filter'] == 'scale':
                return self._scale(original_key, filter_name)
        else:
            raise ParameterNotFound(code=202, msg='Filter type for <%s> has been not present' % filter_name)

    def find_img(self, local_path):
        local_path = os.path.normpath(local_path.lstrip('/'))
        for path_base in current_app.config['IMAGES_PATH']:
            path = os.path.join(current_app.root_path, path_base, local_path)
            if os.path.exists(path):
                return path

    def calculate_size(self, path, **kw):
        return ImageSize(path=self.find_img(path), **kw)

    def _scale(self, original_key, filter_name):
        local_directory = current_app.config['IMAGINE_CACHE'] + '/' + filter_name + \
                          '/' + os.path.dirname(original_key.name)
        makedirs(local_directory)
        local_file_path = current_app.config['IMAGINE_CACHE'] + '/' + filter_name + '/' + original_key.name
        original_key.get_contents_to_filename(local_file_path)

        image = Image.open(local_file_path)
        image = self.resize(image,
                            height=self.filters[filter_name]['height'],
                            mode='fit',
                            width=self.filters[filter_name]['width'])

        format = (os.path.splitext(local_file_path)[1][1:] or 'jpeg').lower()
        format = {'jpg': 'jpeg'}.get(format, format)

        cache_file_path = current_app.config['IMAGINE_CACHE'] + '/' + filter_name + '/' + \
                          os.path.dirname(original_key.name) + '/c_' + os.path.basename(original_key.name)
        cache_file = open(cache_file_path, 'wb')
        image.save(cache_file, format, quality=85)
        cache_file.close()

        cached_key = self.bucket.new_key(current_app.config['IMAGINE_THUMBS_PATH'] + '/' + filter_name + '/' +
                                         original_key.name)
        cached_key.set_contents_from_filename(cache_file_path)
        cached_key.make_public()

        os.remove(local_file_path)
        os.remove(cache_file_path)

        return redirect(cached_key.generate_url(expires_in=0, query_auth=False), code=301)

    def resize(self, image, background=None, **kw):

        size = ImageSize(image=image, **kw)

        # Get into the right colour space.
        if not image.mode.upper().startswith('RGB'):
            image = image.convert('RGBA')

        # Apply any requested transform.
        if size.transform:
            image = Transform(size.transform, image.size).apply(image)

        # Handle the easy cases.
        if size.mode in (modes.RESHAPE, None) or size.req_width is None or size.req_height is None:
            return image.resize((size.width, size.height), Image.ANTIALIAS)

        if size.mode not in (modes.FIT, modes.PAD, modes.CROP):
            raise ValueError('unknown mode %r' % size.mode)

        if image.size != (size.op_width, size.op_height):
            image = image.resize((size.op_width, size.op_height), Image.ANTIALIAS)

        if size.mode == modes.FIT:
            return image

        elif size.mode == modes.PAD:
            pad_color = str(background or 'black')
            padded = Image.new('RGBA', (size.width, size.height), pad_color)
            padded.paste(image, (
                (size.width - size.op_width) // 2,
                (size.height - size.op_height) // 2
            ))
            return padded

        elif size.mode == modes.CROP:

            dx = (size.op_width - size.width) // 2
            dy = (size.op_height - size.height) // 2
            return image.crop(
                (dx, dy, dx + size.width, dy + size.height)
            )

        else:
            raise RuntimeError('unhandled mode %r' % size.mode)


def imagine_filter(path, filter_name, **kwargs):
    self = current_app.extensions['imagine']
    return self.build_url(path, filter_name, **kwargs)




__all__ = [
    'ParameterNotFound', 'FilterNotFound', 'ImagineError', 'OriginalKeyDoesNotExist'
]


class ImagineError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self)
        self.args = args
        self.code = kwargs.pop('code', None)
        self.msg = kwargs.pop('msg', None)

    def __str__(self): # pragma: no cover
        if self.code is not None:
            return '%(code)s: %(msg)s' % self.__dict__
        return Exception.__str__(self)


class ParameterNotFound(ImagineError):
    """
    Parameter not found in config
    """


class FilterNotFound(ImagineError):
    """
    Filter not found in config
    """


class OriginalKeyDoesNotExist(ImagineError):
    """
    Original key does not exist in bucket
    """



try:
    import multiprocessing
except ImportError:
    pass


from setuptools import setup

setup(
    name='Flask-Imagine',
    version='0.1.4',
    description='Dynamic image resizing and storing on AWS S3 for Flask.',
    url='https://github.com/kronas/Flask-Imagine',
        
    author='Kronas',
    author_email='kronas.sw@gmail.com',
    license='MIT',

    packages=['flask_imagine'],

    install_requires=[

        'Flask>=0.9',
        'itsdangerous', # For Flask v0.9
        'Pillow',

    ],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    zip_safe=False,
)
